from pathlib import Path
from typing import Literal
from instaui.version import __version__

_colors = {
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
}


def zero_boot_info(file: Path):
    message = f"""{_app_message("zero")}
{_arrow_right()}file: {_with_color(str(file))}
"""
    print(message)


def web_boot_info(ip: str):
    message = f"""{_app_message("web")}
{_arrow_right()}Local: {_with_color(str(ip), "blue")}
"""
    print(message)


def webview_boot_info():
    pass


def _app_message(mode: str) -> str:
    return f"""{_with_color(f"InstaUI {__version__}")}
mode: {_with_color(mode)}"""


def _with_color(
    text: str, color: Literal["red", "green", "yellow", "blue"] = "green"
) -> str:
    return f"{_colors[color]}{text}\033[0m"


def _arrow_right() -> str:
    return _with_color("➜ ")



# -- __all__
__all__ = [
    "__version__",
    "TMaybeRef",
    "not_",
    "len_",
    "MISSING",
    "TMissing",
    "TBindable",
    "is_bindable",
    "slient",
    "data",
    "TConstData",
    "vue_computed",
    "TVueComputed",
    "js_computed",
    "TJsComputed",
    "computed",
    "TComputed",
    "state",
    "StateModel",
    "local_storage",
    "session_storage",
    "use_dark",
    "PageState",
    "TVForItem",
    "TVForIndex",
    "element_ref",
    "run_element_method",
    "to_json",
    "to_config_data",
    "element",
    "vfor",
    "content",
    "add_style",
    "add_css_link",
    "remove_css_link",
    "add_js_code",
    "add_js_link",
    "add_vue_app_use",
    "use_tailwind",
    "use_page_title",
    "use_language",
    "unwrap_reactive",
    "directive",
    "vif",
    "page",
    "event",
    "js_event",
    "vue_event",
    "event_context",
    "TEventFn",
    "server",
    "context",
    "row",
    "grid",
    "container",
    "text",
    "link",
    "heading",
    "watch",
    "vue_watch",
    "js_watch",
    "js_fn",
    "TWatchState",
    "match",
    "column",
    "experimental",
    "skip_output",
    "str_format",
    "pre_setup_action",
    "patch_set",
    "on_page_init_lifespan",
    "webview",
    "markdown",
    "timer",
    "teleport",
    "icon",
    "box",
    "flex",
    "image",
    "video",
]

# -- static imports
from instaui.version import __version__
from instaui.vars.types import TMaybeRef
from instaui.vars.functions import not_, len_
from instaui.missing import MISSING, TMissing

from instaui.vars.data import const_data as data, TConstData
from instaui.vars.js_computed import js_computed, TJsComputed
from instaui.vars.vue_computed import VueComputed as vue_computed, TVueComputed
from instaui.vars.web_computed import web_computed as computed, TComputed
from instaui.vars.state import state, StateModel
from instaui.vars.local_storage import local_storage
from instaui.vars.session_storage import session_storage
from instaui.vars.use_dark import use_dark
from instaui.vars.use_page_title import use_page_title
from instaui.vars.use_language import use_language
from instaui.vars.unwrap_reactive import unwrap_reactive
from instaui.page_state import PageState
from instaui.vars.vfor_item import TVForItem, TVForIndex
from instaui.vars.element_ref import ElementRef as element_ref, run_element_method
from instaui.html_tools import (
    to_json,
    add_css_link,
    remove_css_link,
    add_js_link,
    add_style,
    add_js_code,
    add_vue_app_use,
    to_config_data,
    use_tailwind,
)

from instaui.components.element import Element as element
from instaui.components.heading import Heading as heading
from instaui.components.directive import Directive as directive
from instaui.components.vfor import VFor as vfor
from instaui.components.vif import VIf as vif
from instaui.components.match import Match as match
from instaui.components.content import Content as content
from instaui.components.timer.timer import Timer as timer
from instaui.components.teleport import teleport
from instaui.components.icon import Icon as icon
from instaui.components.layout.box import Box as box
from instaui.components.layout.flex import (
    Flex as flex,
    FlexRow as row,
    FlexColumn as column,
)
from instaui.components.layout.grid import Grid as grid
from instaui.components.layout.container import Container as container
from instaui.components.text import Text as text
from instaui.components.link import Link as link

from instaui.event.web_event import event, WebEvent as TEventFn
from instaui.event.js_event import js_event
from instaui.event.vue_event import vue_event
from instaui.vars.event_context import EventContext as event_context
from instaui.runtime.context import get_context as context

from instaui.watch.web_watch import watch
from instaui.watch.vue_watch import vue_watch
from instaui.watch.js_watch import js_watch
from instaui.handlers.watch_handler import WatchState as TWatchState
from instaui.skip import skip_output
from instaui.ui_functions.input_slient_data import InputSilentData as slient
import instaui.experimental as experimental

from instaui.ui_functions.server import create_server as server
from instaui.ui_functions.ui_page import page
from instaui.ui_functions.str_format import str_format
from instaui.ui_functions.ui_types import TBindable, is_bindable

from ._events import on_page_init_lifespan
from instaui.extra_libs._import_error import show_error  # noqa: E402, F401
from instaui.js.fn import JsFn as js_fn
from instaui.pre_setup import PreSetupAction as pre_setup_action
from instaui.patch_update import patch_set
from instaui.components.html.image import Image as image
from instaui.components.html.video import Video as video

# -- dynamic imports
from instaui.components.markdown.markdown import Markdown as markdown
from instaui.webview import WebviewWrapper as webview

# -- extra libs



from __future__ import annotations

from typing import Optional
from instaui.common.component_mixin import ContainerMixin, ComponentMixin
from instaui.common.jsonable import Jsonable
from instaui.runtime import get_slot_stacks, pop_slot, get_app_slot, get_current_scope
from instaui.runtime.scope import Scope
from instaui.vars.slot_prop import SlotProp


_DEFAULT_SLOT_NAME = ":"


class SlotManager(Jsonable):
    def __init__(self, *, no_prop=False) -> None:
        super().__init__()
        self._slots: dict[str, Slot] = {}
        self.__no_prop = no_prop

    def get_slot(self, name: str, *, _no_prop: Optional[bool] = None) -> Slot:
        name = normalize_slot_name(name)

        if name not in self._slots:
            no_prop = _no_prop if _no_prop is not None else self.__no_prop
            self._slots[name] = Slot(name, has_children_box=not no_prop)

        return self._slots[name]

    @property
    def default(self):
        return self.get_slot(_DEFAULT_SLOT_NAME)

    def _to_json_dict(self):
        return {name: slot._to_json_dict() for name, slot in self._slots.items()}

    def has_slot(self) -> bool:
        return len(self._slots) > 0 and any(
            slot.has_items() for slot in self._slots.values()
        )


class Slot(ContainerMixin, Jsonable):
    def __init__(self, name: str, *, has_children_box=True) -> None:
        super().__init__()

        self._id: Optional[str] = None
        self._name = name
        self._children: list[ComponentMixin] = []
        self._slot_props_used_id: Optional[int] = None
        self.__has_children_box = has_children_box

        if self.__has_children_box:
            self._scope = _new_scope_slot()

        self.__define_scope = (
            self._scope if self.__has_children_box else get_current_scope()
        )

        self._used_prop: Optional[SlotProp] = None

    @property
    def _define_scope(self):
        return self.__define_scope

    def has_items(self):
        if self.__has_children_box:
            return len(self._scope._items) > 0

        return len(self._children) > 0

    def _mark_binding_used(self, var_id: int):
        self._slot_props_used_id = var_id

    def slot_props(self, name: str):
        if not self.__has_children_box:
            raise ValueError(
                "Not allowed to use slot props when slot has no children box"
            )

        self._use_slot_props = True

        if not self._used_prop:
            self._used_prop = SlotProp(name, slot=self)

        return self._used_prop[name]

    def __getitem__(self, item: str):
        return self.slot_props(item)

    def __enter__(self):
        if self.__has_children_box:
            self._scope.__enter__()
            return self

        get_slot_stacks().append(self)
        return self

    def __exit__(self, *_):
        if self.__has_children_box:
            self._scope.__exit__(*_)
            return

        pop_slot()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self.__has_children_box:
            data["scope"] = self._scope

        if self._children:
            data["items"] = self._children

        if self._slot_props_used_id:
            data["usePropId"] = self._slot_props_used_id

        return data

    def add_item(self, item: ComponentMixin):
        self._children.append(item)


def _new_scope_slot():
    app = get_app_slot()
    _new_scope = Scope(app.gen_scope_id(), add_to_container=False)

    return _new_scope


def normalize_slot_name(name: str):
    return _DEFAULT_SLOT_NAME if name == "default" else name



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Sequence
from instaui.common.binding_track_mixin import (
    mark_binding,
    try_mark_binding,
    is_binding_tracker,
)

from instaui.vars._types import InputBindingType, OutputSetType

if TYPE_CHECKING:
    from instaui.runtime.scope import Scope


class CanInputMixin(ABC):
    @abstractmethod
    def _to_event_input_type(self) -> InputBindingType:
        pass


class CanOutputMixin(ABC):
    @abstractmethod
    def _to_event_output_type(self) -> OutputSetType:
        pass


def inputs_to_config(inputs: Sequence[CanInputMixin]):
    return [
        {
            "value": mark_binding(input) if is_binding_tracker(input) else input,
            "type": input._to_event_input_type().value
            if isinstance(input, CanInputMixin)
            else InputBindingType.Data.value,
        }
        for input in inputs
    ]


def outputs_to_config(
    outputs: Sequence[CanOutputMixin], *, scope: Optional[Scope] = None
):
    return [
        {
            "ref": try_mark_binding(ref, scope=scope),
            "type": ref._to_event_output_type().value,
        }
        for ref in outputs
    ]


def _assert_outputs_be_can_output_mixin(outputs: Sequence):
    for output in outputs:
        if not isinstance(output, CanOutputMixin):
            raise TypeError("The outputs parameter must be a `ui.state`")



from instaui.runtime import get_app_slot
from instaui.handlers import watch_handler
from instaui.handlers import event_handler
from instaui.response import response_data


class Api:
    def watch_call(self, data: dict):
        hkey = data.pop("key")
        handler_info = watch_handler.get_handler_info(hkey)
        if handler_info is None:
            return {"error": "watch handler not found"}

        update_app_page_info(data)

        result = handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return response_data(handler_info.outputs_binding_count, result)

    def event_call(self, data: dict):
        handler = event_handler.get_handler(data["hKey"])
        if handler is None:
            raise ValueError("event handler not found")

        update_app_page_info(data)

        args = [bind for bind in data.get("bind", [])]

        result = handler.fn(*handler.get_handler_args(args))
        return response_data(handler.outputs_binding_count, result)


def update_app_page_info(data: dict):
    app = get_app_slot()

    page_info = data.get("page", {})
    app._page_path = page_info["path"]

    if "params" in page_info:
        app._page_params = page_info["params"]

    if "queryParams" in page_info:
        app._query_params = page_info["queryParams"]


def _get_binds_from_data(data: dict):
    return data.get("input", [])



from __future__ import annotations
from typing import Any, Callable
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from instaui.runtime import new_app_slot, reset_app_slot
from instaui.runtime.utils import init_base_scope
from .request_context import set_current_request


class RequestContextMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: Callable) -> Any:
        set_current_request(request)
        app, system_slot_token = new_app_slot("web")
        init_base_scope(app)

        try:
            response = await call_next(request)
        finally:
            reset_app_slot(system_slot_token)

        return response


class NoCacheDebugModeMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: Callable) -> Any:
        response = await call_next(request)

        if request.url.path.endswith((".js", ".css")):
            response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
            response.headers["Pragma"] = "no-cache"
            response.headers["Expires"] = "0"

        return response



from __future__ import annotations
import typing
from instaui.common.jsonable import Jsonable
from instaui.runtime import try_new_scope_on_ui_state
from instaui.vars._types import InputBindingType
from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper
from instaui.common.var_track_mixin import VarTrackerMixin, VarTrackerHelper
from instaui.common.binding_track_mixin import (
    BindingTrackerMixin,
    BindingTrackerHelper,
    try_mark_binding,
)

if typing.TYPE_CHECKING:
    from instaui.runtime.scope import Scope


_R_TYPE = typing.TypeVar("_R_TYPE")


class JsComputed(
    Jsonable,
    PathVar,
    ObservableMixin,
    CanInputMixin,
    StrFormatBindingMixin,
    VarTrackerMixin,
    BindingTrackerMixin,
    ElementBindingMixin,
):
    def __init__(
        self,
        *,
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        code: str,
        async_init_value: typing.Optional[typing.Any] = None,
        deep_compare_on_input: bool = False,
        _tool_name: typing.Optional[str] = None,
    ) -> None:
        self.code = code
        self._tool_name = _tool_name
        self._org_inputs = inputs or []

        self._define_scope = try_new_scope_on_ui_state()

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )

        self._async_init_value = async_init_value
        self._deep_compare_on_input = deep_compare_on_input

        def upstreams_binding_callback(scope: Scope):
            self._inputs = [
                try_mark_binding(input, scope=scope) for input in self._inputs
            ]

        self.__binding_tracker = BindingTrackerHelper(
            define_scope=self._define_scope,
            upstreams_callback=upstreams_binding_callback,
        )

        self.__var_tracker = VarTrackerHelper(
            var_id_gen_fn=lambda: self._define_scope.register_js_computed(self),
            upstreams_getter=lambda: [inputs],
        )

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["id"] = self.__var_tracker.var_id

        if self._inputs:
            data["inputs"] = self._inputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        if self._async_init_value is not None:
            data["asyncInit"] = self._async_init_value

        if self._deep_compare_on_input is not False:
            data["deepEqOnInput"] = 1
        if self._tool_name is not None:
            data["tool"] = self._tool_name

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _mark_as_used(self):
        self.__var_tracker.mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        return self.__binding_tracker.mark_binding(
            var_id=self.__var_tracker.var_id, scope=scope
        )


TJsComputed = JsComputed


@typing.overload
def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
) -> typing.Any: ...


@typing.overload
def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
    r_type: typing.Optional[typing.Type[_R_TYPE]] = None,
) -> _R_TYPE: ...


def js_computed(
    *,
    inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
    code: str,
    async_init_value: typing.Optional[typing.Any] = None,
    deep_compare_on_input: bool = False,
    r_type: typing.Optional[typing.Type[_R_TYPE]] = None,
) -> typing.Union[JsComputed, _R_TYPE]:
    """
    A client-side computed property that evaluates JavaScript code to derive reactive values.

    Args:
        inputs (typing.Optional[typing.Sequence[TObservableInput]], optional): Reactive data sources to monitor.
                                                  Changes to these values trigger re-computation.
        code (str): JavaScript code to execute for computing the value.
        async_init_value (typing.Optional[typing.Any], optional): Initial value to use before first successful async evaluation.

    # Example:
    .. code-block:: python
        a = ui.state(0)

        plus_one = ui.js_computed(inputs=[a], code="a=> a+1")

        html.number(a)
        ui.text(plus_one)
    """

    jc = JsComputed(
        inputs=inputs,
        code=code,
        async_init_value=async_init_value,
        deep_compare_on_input=deep_compare_on_input,
    )

    if r_type is None:
        return jc

    return typing.cast(_R_TYPE, jc)



from abc import ABC, abstractmethod
import typing

from instaui.event.event_modifier import TEventModifier


class EventMixin(ABC):
    @abstractmethod
    def event_type(self) -> typing.Literal["web", "js"]:
        pass

    @abstractmethod
    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]],
    ) -> dict:
        pass



from __future__ import annotations
from typing import Optional, TypeVar

from .ref_base import RefBase

_T_Value = TypeVar("_T_Value")


class Ref(RefBase):
    def __init__(
        self,
        value: Optional[_T_Value] = None,
        deep_compare: bool = False,
    ) -> None:
        super().__init__(value=value, deep_compare=deep_compare)



from instaui.vars._types import OutputSetType
from instaui.vars.mixin_types.py_binding import CanOutputMixin


class JsOutput(CanOutputMixin):
    def _to_output_config(self):
        return {"type": "js"}

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.JsCode



from __future__ import annotations
from instaui.common.binding_track_mixin import try_mark_binding
from instaui.common.component_mixin import ComponentMixin
from instaui.common.var_track_mixin import mark_as_used
from instaui.vars.types import TMaybeRef
from instaui.components.logic_component import LogicContainer


class VIf(LogicContainer):
    def __init__(self, on: TMaybeRef[bool]):
        """
        A conditional container that renders its children only when the given condition is True.

        Args:
            on (TMaybeRef[bool]): A boolean or reactive reference that determines whether the container's
                                child elements should be displayed. If True, children are rendered;
                                if False, children are hidden.


        Example:
        .. code-block:: python
            from instaui import ui, html

            value = ui.state(False)

            html.button("toggle").on_click(
                ui.js_event(inputs=[value], outputs=[value], code="(v)=> !v")
            )

            with ui.vif(value):
                ui.text("show")
        """

        super().__init__("vif")
        mark_as_used(on)
        self._on = try_mark_binding(on)
        self._items: list[ComponentMixin] = []

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["on"] = self._on

        if self._items:
            data["items"] = self._items

        return data

    def add_item(self, item: ComponentMixin):
        self._items.append(item)



from typing import Any, Tuple, TypeVar, Union, cast
from instaui.common.const_data_mixin import ConstDataMixin
from instaui.runtime.scope import Scope
from instaui.vars.path_var import PathVar
from instaui.vars.ref import Ref
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.common.var_track_mixin import VarTrackerMixin
from instaui.common.binding_track_mixin import BindingTrackerMixin

from instaui.common.jsonable import Jsonable
from instaui.vars.ref_base import RefBase


from pydantic import BaseModel, RootModel


_T = TypeVar("_T")


_ProxyModel = RootModel


class RefProxy(
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    VarTrackerMixin,
    BindingTrackerMixin,
    PathVar,
    Jsonable,
):
    def __init__(self, data, ref_base: RefBase) -> None:
        self._ref_ = ref_base
        self._prop_names_ = set(data.keys()) if isinstance(data, dict) else set()

    @property
    def __ref_(self):
        return super().__getattribute__("_ref_")

    def __deepcopy__(self, memo):
        return self.__ref_.__deepcopy__(memo)

    def __getattribute__(self, name):
        if name not in super().__getattribute__("_prop_names_"):
            return super().__getattribute__(name)

        return self.__ref_[name]

    def __getitem__(self, name):
        return self.__ref_[name]

    def not_(self):
        return self.__ref_.not_()

    def __add__(self, other: str):
        return self.__ref_ + other

    def __radd__(self, other: str):
        return other + self.__ref_

    def __sub__(self, other: Union[int, float]):
        return self.__ref_ - other

    def __rsub__(self, other: Union[int, float]):
        return other - self.__ref_

    def __mul__(self, other: Union[int, float]):
        return self.__ref_ * other

    def __rmul__(self, other: Union[int, float]):
        return other * self.__ref_

    def __truediv__(self, other: Union[int, float]):
        return self.__ref_ / other

    def __rtruediv__(self, other: Union[int, float]):
        return other / self.__ref_

    def __and__(self, other: Any):
        return self.__ref_ & other

    def __or__(self, other: Any):
        return self.__ref_ | other

    def __lt__(self, other):
        return self.__ref_ < other

    def __le__(self, other):
        return self.__ref_ <= other

    def __gt__(self, other):
        return self.__ref_ > other

    def __ge__(self, other):
        return self.__ref_ >= other

    def __ne__(self, other):
        return self.__ref_ != other

    def len_(self):
        return self.__ref_.len_()

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return self.__ref_._to_str_format_binding(order)

    def _to_json_dict(self):
        return self.__ref_._to_json_dict()

    def _to_event_output_type(self) -> OutputSetType:
        return self.__ref_._to_event_output_type()

    def _to_event_input_type(self) -> InputBindingType:
        return self.__ref_._to_event_input_type()

    def _mark_as_used(self):
        self._ref_._mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        return self._ref_._mark_binding(scope)


class StateModel(BaseModel, Jsonable):
    pass

    def _to_json_dict(self):
        return self.model_dump()


def state(
    value: _T,
    deep_compare: bool = False,
) -> _T:
    """
    Creates a reactive state wrapper that enables automatic UI updates when its value or
    nested elements change. Supports primitives, lists, dictionaries, and provides
    reactive operations such as indexing, comparisons, and length evaluation.

    Args:
        value (_T): The initial value to be wrapped as a reactive state. Can be any
            Python type including primitives, lists, or dictionaries.
        deep_compare (bool): Whether state updates should compare nested values to
            determine change propagation. Defaults to False.

    Example:
    .. code-block:: python
        from instaui import ui, html

        # Value reactivity
        s = ui.state('hello')
        html.input(s)
        ui.text(s)


        # List reactivity
        items = ui.state([1, 2, 3])
        html.number(items[0])
        html.ul.from_list(items)


        # Dictionary mutation triggers UI updates
        data = ui.state({"name": "John"})
        html.input(data["name"])


        # State used with js_computed
        data = ui.state({"items": [1, 2, 3]})
        items = ui.js_computed(inputs=[data], code="(d)=> d.items")
        html.ul.from_list(items)


        # Boolean operations maintain reactivity
        value = ui.state([{"state": True}])
        html.checkbox(value[0]["state"])
        html.paragraph(ui.not_(value[0]["state"]))


        # String or list length as reactive value
        text = ui.state("Hello")
        html.paragraph(ui.str_format("len: {}", ui.len_(text)))


        # Comparison operators trigger updates
        a = ui.state(1)
        b = ui.state(2)
        html.paragraph(a < b)  # UI updates when a or b changes
    """

    if isinstance(value, RefProxy):
        return value

    new_value = (
        value.get_value()
        if isinstance(value, ConstDataMixin)
        else _ProxyModel(value).model_dump()
    )
    obj = RefProxy(new_value, Ref(value, deep_compare=deep_compare))
    return cast(_T, obj)



from __future__ import annotations
from typing import Any, Optional, Union
from instaui.components.element import Element
from instaui.components.value_element import ValueElement
from instaui.components.mixins import CanDisabledMixin
from ._mixins import InputEventMixin


class Input(InputEventMixin, CanDisabledMixin, ValueElement[str]):
    def __init__(
        self,
        value: Union[str, Any, None] = None,
        *,
        model_value: Union[str, Any, None] = None,
        disabled: Optional[bool] = None,
    ):
        super().__init__("input", value, is_html_component=True)

        if disabled is not None:
            self.props({"disabled": disabled})
        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self



from dataclasses import dataclass
import itertools
from pathlib import Path
import shutil
from typing import Callable, Optional
from urllib.parse import quote

from instaui.version import __version__ as _INSTAUI_VERSION
from instaui.html_tools import to_config_data
from instaui.runtime import get_app_slot, get_default_app_slot
from instaui.systems import file_system
from instaui.template import webview_template
from instaui import consts
from instaui.runtime.resource import StyleTag
from . import _utils


_INDEX_HTML_NAME = "index.html"


@dataclass
class ResourceInfo:
    title: str
    index_html_url: str


def resource_to_assets(
    page_url: str, assets_path: Path, on_app_mounted: Optional[Callable] = None
):
    page_dir = assets_path.joinpath(quote(page_url, safe=""))
    _utils.reset_dir(page_dir)

    relative_to_assets = _get_relative_to_assets(page_dir)
    file_to_assets = _get_file_to_assets(page_dir)

    config_data = to_config_data()

    system_slot = get_app_slot()
    default_app_slot = get_default_app_slot()
    html_resource = system_slot._html_resource
    default_html_resource = default_app_slot._html_resource

    if html_resource.favicon:
        favicon_url = file_to_assets(html_resource.favicon)
    else:
        if default_html_resource.favicon:
            favicon_url = file_to_assets(default_html_resource.favicon)
        else:
            favicon_url = file_to_assets(consts.FAVICON_PATH)

    model = webview_template.WebViewTemplateModel(
        version=_INSTAUI_VERSION,
        vue_js_code=file_to_assets(consts.VUE_ES_JS_PATH),
        instaui_js_code=file_to_assets(consts.APP_ES_JS_PATH),
        css_links=[
            file_to_assets(consts.APP_CSS_PATH),
        ],
        config_dict=config_data,
        favicon_url=favicon_url,
        title=html_resource.title or default_html_resource.title or consts.PAGE_TITLE,
        on_app_mounted=on_app_mounted,
    )

    # register custom components
    for component in system_slot._component_dependencies:
        if not component.esm:
            continue

        component_url = file_to_assets(component.esm)
        model.add_extra_import_map(component_url, file_to_assets(component.esm))

        model.vue_app_component.append(
            webview_template.WebViewVueAppComponent(
                name=component.tag_name,
                url=component_url,
            )
        )

        if component.css:
            for css_link in component.css:
                css_resource = file_to_assets(css_link)
                if css_link.is_file():
                    model.css_links.append(css_resource)

        if component.externals:
            for name, file in component.externals.items():
                model.add_extra_import_map(name, file_to_assets(file))

    # register custom plugins
    for plugin in set(
        itertools.chain(
            system_slot._plugin_dependencies, default_app_slot._plugin_dependencies
        )
    ):
        if not plugin.esm:
            continue

        model.vue_app_use.append(plugin.name)

        model.add_extra_import_map(plugin.name, file_to_assets(plugin.esm))

        for name, file in plugin.externals.items():
            model.add_extra_import_map(name, file_to_assets(file))

        for css_link in plugin.css:
            model.css_links.append(file_to_assets(css_link.file))

    # css file link to web static link
    for link in html_resource.get_valid_css_links(
        default_html_resource._css_links_manager
    ):
        if isinstance(link, Path):
            model.css_links.append(file_to_assets(link))
        else:
            model.css_links.append(link)

    # js file link to web static link
    for info in html_resource.get_valid_js_links(
        default_html_resource._js_links_manager
    ):
        link = file_to_assets(info.link) if isinstance(info.link, Path) else info.link
        model.js_links.append(link)

    for js_code in itertools.chain(
        html_resource._script_tags, default_html_resource._script_tags
    ):
        model.script_tags.append(js_code)

    for sylte_tag in StyleTag.merge_by_group_id(
        itertools.chain(html_resource._style_tags, default_html_resource._style_tags)
    ):
        model.style_tags.append(sylte_tag)

    html_str = webview_template.render_wbeview_html(model)

    index_html_path = page_dir / _INDEX_HTML_NAME
    index_html_path.write_text(html_str, encoding="utf-8")
    relative_to_assets(index_html_path)
    index_html_url = str(index_html_path.absolute())
    return ResourceInfo(
        title="test",
        index_html_url=index_html_url,
    )


def _get_relative_to_assets(assets_path: Path):
    def wrapper(file_path: Path, relative_parent=False):
        return str(
            file_path.relative_to(
                assets_path.parent if relative_parent else assets_path
            )
        ).replace("\\", "/")

    return wrapper


def _get_file_to_assets(assets_path: Path):
    relative_to_assets = _get_relative_to_assets(assets_path)

    def wrapper(file_path: Path):
        hash_part = file_system.generate_hash_name_from_path(file_path.parent)
        new_folder_path = assets_path.joinpath(hash_part)

        if not new_folder_path.exists():
            new_folder_path.mkdir(parents=True)

        new_path = new_folder_path.joinpath(file_path.name)

        if file_path.is_file():
            shutil.copyfile(file_path, new_path)
            return "./" + relative_to_assets(new_path)
        else:
            shutil.copytree(file_path, new_path, dirs_exist_ok=True)
            return "./" + relative_to_assets(new_path) + "/"

    return wrapper



from pathlib import Path
from typing import Union
import inspect


def get_caller_path(path_or_str: Union[str, Path], up: int = 1) -> Path:
    if isinstance(path_or_str, Path):
        return path_or_str

    frame = inspect.currentframe()
    try:
        # (+1 because the current function itself counts as one frame)
        for _ in range(up + 1):
            if frame is None:
                raise RuntimeError("Frame stack too shallow")
            frame = frame.f_back

        assert frame is not None, "Frame stack too shallow"

        caller_file = Path(inspect.getfile(frame)).resolve()
        caller_dir = caller_file.parent

        return (caller_dir / path_or_str).resolve()
    finally:
        del frame



from __future__ import annotations
from typing import Union
from instaui.common.binding_track_mixin import BindingTrackerMixin
from instaui.components.element import Element
from instaui.components.value_element import ValueElement
from instaui import consts
from instaui.vars.types import TMaybeRef
from ._mixins import InputEventMixin

_T_value = Union[int, float]


class Range(InputEventMixin, ValueElement[_T_value]):
    def __init__(
        self,
        value: Union[_T_value, TMaybeRef[_T_value], None] = None,
        *,
        min: Union[_T_value, TMaybeRef[_T_value], None] = None,
        max: Union[_T_value, TMaybeRef[_T_value], None] = None,
        step: Union[_T_value, TMaybeRef[_T_value], None] = None,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "range"})

        if min is not None:
            self.props({"min": min})
        if max is not None:
            self.props({"max": max})
        if step is not None:
            self.props({"step": step})

    def vmodel(
        self,
        value: BindingTrackerMixin,
        modifiers: Union[consts.TModifier, list[consts.TModifier], None] = None,
        *,
        prop_name: str = "value",
    ):
        modifiers = modifiers or []
        if isinstance(modifiers, str):
            modifiers = [modifiers]

        modifiers_with_number = list(set([*modifiers, "number"]))

        return super().vmodel(value, modifiers_with_number, prop_name=prop_name)  # type: ignore

    def _input_event_mixin_element(self) -> Element:
        return self



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Optional, Set
from contextlib import contextmanager
from instaui.common.jsonable import Jsonable
from instaui.common.component_mixin import ContainerMixin, ComponentMixin
from .resource import HtmlResource
from instaui.consts import _T_App_Mode
from types import MappingProxyType

if TYPE_CHECKING:
    from instaui.runtime.scope import Scope, BaseScope
    from instaui.dependencies.component_dependency import ComponentDependencyInfo
    from instaui.dependencies.plugin_dependency import PluginDependencyInfo
    from instaui.spa_router._route_model import RouteCollector


class App(Jsonable):
    _default_app_slot: ClassVar[Optional[App]] = None
    _web_server_info: ClassVar[Optional[dict]] = None

    def __init__(
        self,
        *,
        mode: _T_App_Mode,
        debug_mode: bool = False,
        meta: Optional[dict] = None,
    ) -> None:
        super().__init__()
        self._scope_id_counter = 0
        self._scoped_style_id_counter = 0
        self._mode: _T_App_Mode = mode
        self.meta = meta
        self.__debug_mode = debug_mode

        self._slots_stacks: list[ContainerMixin] = []
        self._base_scope: Optional[BaseScope] = None

        self._scope_stack: list[Scope] = []
        self._html_resource = HtmlResource()
        self._component_dependencies: Set[ComponentDependencyInfo] = set()
        self._temp_component_dependencies: dict[str, ComponentDependencyInfo] = {}
        self._plugin_dependencies: Set[PluginDependencyInfo] = set()

        self._page_path: Optional[str] = None
        self._page_params: dict[str, Any] = {}
        self._query_params: dict[str, Any] = {}
        self._route_collector: Optional[RouteCollector] = None
        self._router_base_scope: Optional[Scope] = None
        self._used_icons: set[str] = set()
        self._used_icons_set: set[str] = set()

    def collect_icon(self, icon_name: str):
        self._used_icons.add(icon_name)

    def collect_icon_set(self, icon_set_name: str):
        self._used_icons_set.add(icon_set_name)

    @contextmanager
    def _mark_router_base_scope(self, scope: Scope):
        self._router_base_scope = scope
        yield
        self._router_base_scope = None

    @property
    def debug_mode(self) -> bool:
        return self.__debug_mode

    @property
    def router_base_scope(self):
        return self._router_base_scope

    def setup(self, base_scope: BaseScope):
        base_scope.__enter__()
        self._base_scope = base_scope

    @property
    def mode(self) -> _T_App_Mode:
        return self._mode

    @property
    def base_scope(self) -> BaseScope:
        return self._base_scope  # type: ignore

    def gen_scope_id(self) -> int:
        sid = self._scope_id_counter
        self._scope_id_counter += 1
        return sid

    @property
    def top_scope(self) -> Scope:
        return self._scope_stack[0]

    @property
    def page_path(self) -> str:
        assert self._page_path is not None, "Page path is not set"
        return self._page_path  # type: ignore

    @property
    def page_params(self):
        return MappingProxyType(self._page_params)

    @property
    def query_params(self):
        return MappingProxyType(self._query_params)

    def gen_scoped_style_group_id(self):
        gid = f"scoped-style-{self._scoped_style_id_counter}"
        self._scoped_style_id_counter += 1
        return gid

    def reset_html_resource(self):
        self._html_resource = HtmlResource()

    def add_temp_component_dependency(self, dependency: ComponentDependencyInfo):
        self._temp_component_dependencies[dependency.tag_name] = dependency

    def get_temp_component_dependency(
        self, tag_name: str, default: ComponentDependencyInfo
    ) -> ComponentDependencyInfo:
        return self._temp_component_dependencies.get(tag_name, default)

    def has_temp_component_dependency(self, tag_name: str):
        return tag_name in self._temp_component_dependencies

    def use_component_dependency(
        self, dependency: ComponentDependencyInfo, *, replace=False
    ) -> None:
        if replace:
            self._component_dependencies.discard(dependency)

        self._component_dependencies.add(dependency)

    def use_plugin_dependency(self, dependency: PluginDependencyInfo) -> None:
        self._plugin_dependencies.add(dependency)

    def register_router(self, collector: RouteCollector) -> None:
        self._route_collector = collector

    def append_component_to_container(self, component: ComponentMixin):
        self._slots_stacks[-1].add_item(component)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._page_path:
            url_info = {"path": self.page_path}
            if self._page_params:
                url_info["params"] = self._page_params  # type: ignore

            data["url"] = url_info

        if self._route_collector is not None:
            data["router"] = self._route_collector

        if self._web_server_info is not None:
            data["webInfo"] = self._web_server_info

        data["scope"] = self._base_scope

        icons = {}
        if self._used_icons:
            icons["names"] = list(self._used_icons)

        if self._used_icons_set:
            icons["sets"] = list(self._used_icons_set)

        if icons:
            data["icons"] = icons

        return data


class DefaultApp(App):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DefaultApp, cls).__new__(cls)
        return cls._instance

    def append_component_to_container(self, component: ComponentMixin):
        raise ValueError("Operations are not allowed outside of ui.page")

    @property
    def base_scope(self) -> Scope:
        raise ValueError("Operations are not allowed outside of ui.page")



from __future__ import annotations
from typing import (
    Optional,
)
from instaui.components.element import Element

from instaui.vars.types import TMaybeRef


class Link(Element):
    def __init__(
        self,
        href: Optional[TMaybeRef[str]] = None,
        *,
        text: Optional[TMaybeRef[str]] = None,
    ):
        super().__init__("a")

        if text is not None:
            self.props(
                {
                    "innerText": text,
                }
            )

        if href is not None:
            self.props(
                {
                    "href": href,
                }
            )



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Optional, Union
from instaui.components.element import Element
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from . import _src_utils

if TYPE_CHECKING:
    from instaui.vars.types import TMaybeRef


class Image(Element):
    def __init__(
        self,
        src: Optional[TMaybeRef[str]] = None,
        *,
        alt: Optional[TMaybeRef[str]] = None,
        width: Optional[TMaybeRef[Union[int, str]]] = None,
        height: Optional[TMaybeRef[Union[int, str]]] = None,
        title: Optional[TMaybeRef[str]] = None,
        srcset: Optional[TMaybeRef[str]] = None,
        sizes: Optional[TMaybeRef[str]] = None,
        loading: Optional[TMaybeRef[Literal["lazy", "eager"]]] = None,
    ):
        """
        Creates an HTML `<img>` element with configurable attributes.

        The `src` parameter specifies the image source. If a relative path is provided
        (starting with `/`), it resolves to files in the application's `assets/` directory.
        Absolute URLs are used directly.

        Args:
            src: Path to the image file or URL. Relative paths must start with `/` and
                 resolve to files in the `assets/` directory. Example: `/xxx.png` refers to
                 `assets/xxx.png`.
            alt: Alternative text for accessibility and SEO when the image cannot be displayed.
            width: Width of the image in pixels or percentage (e.g., `300` or `"100%"`).
            height: Height of the image in pixels or percentage (e.g., `200` or `"50%"`).
            title: Tooltip text displayed when hovering over the image.
            srcset: Comma-separated list of image URLs with descriptors for responsive design.
            sizes: Media query conditions to define image size based on viewport width.
            loading: Specifies whether the image should load lazily (`"lazy"`) or eagerly (`"eager"`).

        Example:
        .. code-block:: python
            # Renders <img src="/xxx.png" alt="Example Image" width="300" height="200">
            html.image("/xxx.png", alt="Example Image", width=300, height=200)
        """
        super().__init__("img")

        if src is not None:
            if isinstance(src, ElementBindingMixin):
                src = _src_utils.complete_src_computed(src)
            else:
                src = _src_utils.complete_src(src)

        self.props(
            {
                "src": src,
                "alt": alt,
                "width": width,
                "height": height,
                "title": title,
                "srcset": srcset,
                "sizes": sizes,
                "loading": loading,
            }
        )



from __future__ import annotations
from typing import TYPE_CHECKING


from instaui.runtime.scope import BaseScope
from instaui.vars.use_dark import UseDarkRef
from instaui.vars.use_page_title import UsePageTitleRef
from instaui.vars.use_language import UseLanguageRef

if TYPE_CHECKING:
    from instaui.runtime._app import App


def init_base_scope(app: App):
    bs = BaseScope(app.gen_scope_id())
    app.setup(bs)
    bs.setup(
        use_dark=UseDarkRef(),
        use_page_title=UsePageTitleRef(),
        use_language=UseLanguageRef(),
    )



import typing
from instaui.common.binding_track_mixin import (
    BindingTrackerMixin,
    try_mark_binding,
    is_binding_tracker,
)
from instaui.common.var_track_mixin import VarTrackerMixin, mark_as_used
from instaui.runtime.scope import Scope
from instaui.vars._types import InputBindingType
from instaui.vars.mixin_types.py_binding import CanInputMixin


class InputSilentData(CanInputMixin, BindingTrackerMixin, VarTrackerMixin):
    def __init__(self, value: typing.Union[BindingTrackerMixin, typing.Any]) -> None:
        """
        Wraps a value to mark it as "silent" for reactive tracking,
        so that changes to this value do not trigger dependent watchers
        or computed functions.

        Args:
            value (Union[BindingTrackerMixin, Any]): The value to wrap.
                Can be any data type or a reactive binding,
                which will be excluded from triggering reactive updates.


        Example:
        .. code-block:: python
            from instaui import ui, html

            a = ui.state("a")
            b = ui.state("b")
            result = ui.state("result")

            # Use silent data for 'b' so changes to 'b' don't trigger recomputation
            @ui.watch(inputs=[a, ui.slient(b)], outputs=[result])
            def only_a_changed(a: str, b: str):
                return f"{a}+{b}"

            # Only changes to 'a' will update the result
            html.input(a).classes("a")
            html.input(b).classes("b")
            html.paragraph(result)
        """

        self.value = value

    def is_const_value(self) -> bool:
        return not is_binding_tracker(self.value)

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _mark_binding(self, scope: Scope) -> dict:
        return try_mark_binding(self.value, scope=scope)

    def _mark_as_used(self):
        mark_as_used(self.value)



from __future__ import annotations
import itertools
from pathlib import Path
from typing import Optional
from instaui.version import __version__ as _INSTA_VERSION
import instaui.consts as consts
from instaui.runtime import get_app_slot, get_default_app_slot
from instaui.template import render_zero_html
from instaui.template import zero_template
from instaui.html_tools import to_config_data
from instaui.runtime.dataclass import JsLink
from instaui.runtime.resource import HtmlResource, StyleTag
from .options import ZeroOptions


def to_html(
    file: Path,
    *,
    options: Optional[ZeroOptions] = None,
    use_empty_html_resource: bool = False,
):
    file = Path(file)

    raw = to_html_str(options, use_empty_html_resource)
    file.write_text(raw, "utf8")

    return file.resolve().absolute()


def get_template_model(options: ZeroOptions, use_empty_html_resource: bool = False):
    system_slot = get_app_slot()

    default_app_slot = get_default_app_slot()
    html_resource = system_slot._html_resource
    default_html_resource = (
        _empty_html_resource()
        if use_empty_html_resource
        else default_app_slot._html_resource
    )

    config_data = to_config_data()

    model = zero_template.ZeroTemplateModel(
        import_maps_cdn_overrides=options.get_import_maps_cdn_overrides(),
        css_links_cdn_overrides=options.get_css_links_cdn_overrides(),
        version=_INSTA_VERSION,
        icons_svg_code=options.icons_svg_content,
        vue_js_code=consts.VUE_ES_JS_PATH,
        instaui_js_code=consts.APP_ES_JS_PATH,
        css_links=[
            consts.APP_CSS_PATH,
        ],
        config_dict=config_data,
        favicon=html_resource.favicon
        or default_html_resource.favicon
        or consts.FAVICON_PATH,
        title=html_resource.title or default_html_resource.title or consts.PAGE_TITLE,
    )

    # register custom components
    for component in system_slot._component_dependencies:
        if not component.esm:
            continue

        model.vue_app_component.append(
            zero_template.ZeroVueAppComponent(
                name=component.tag_name,
                url=component.esm,
            )
        )

        if component.css:
            for css_link in component.css:
                model.add_css_link(css_link)

        if component.externals:
            for name, url in component.externals.items():
                if url.is_file():
                    model.add_extra_import_map(name, url)

    # register custom plugins
    for plugin in set(
        itertools.chain(
            system_slot._plugin_dependencies, default_app_slot._plugin_dependencies
        )
    ):
        if not plugin.esm:
            continue

        model.vue_app_use.append(plugin.name)

        model.add_extra_import_map(plugin.name, plugin.esm)

        for name, url in plugin.externals.items():
            model.add_extra_import_map(name, url)

        for css in plugin.css:
            model.add_css_link(css)

    # css file link to web static link
    for link in html_resource.get_valid_css_links(
        default_html_resource._css_links_manager
    ):
        if isinstance(link, Path):
            model.add_css_link(link)

    # js file link to web static link
    for info in html_resource.get_valid_js_links(
        default_html_resource._js_links_manager
    ):
        if isinstance(info.link, Path):
            model.js_links.append(JsLink(info.link))

    for js_code in itertools.chain(
        html_resource._script_tags, default_html_resource._script_tags
    ):
        model.script_tags.append(js_code)

    for sylte_tag in StyleTag.merge_by_group_id(
        itertools.chain(html_resource._style_tags, default_html_resource._style_tags)
    ):
        model.style_tags.append(sylte_tag)

    return model


def to_html_str(
    options: Optional[ZeroOptions] = None, use_empty_html_resource: bool = False
):
    model = get_template_model(options or ZeroOptions(), use_empty_html_resource)
    return render_zero_html(model)


def _empty_html_resource():
    return HtmlResource()



from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from instaui.runtime import get_app_slot
from instaui.vars._types import OutputSetType
from instaui.vars.mixin_types.py_binding import CanOutputMixin
from pydantic import BaseModel
from instaui.common.var_track_mixin import VarTrackerMixin, VarTrackerHelper
from instaui.common.binding_track_mixin import BindingTrackerMixin, BindingTrackerHelper

if TYPE_CHECKING:
    from instaui.runtime.scope import Scope


class RouterActionVar(BindingTrackerMixin, VarTrackerMixin, CanOutputMixin):
    def __init__(self):
        base_scope = get_app_slot().base_scope
        assert base_scope is not None, "Router base scope not found"
        self._define_scope = base_scope

        def var_id_gen_fn():
            var_id = self._define_scope.register_router_action_used()
            return var_id

        self.__var_tracker = VarTrackerHelper(var_id_gen_fn=var_id_gen_fn)
        self.__binding_tracker = BindingTrackerHelper(define_scope=self._define_scope)

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.RouterAction

    def _mark_as_used(self):
        self.__var_tracker.mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        return self.__binding_tracker.mark_binding(
            var_id=self.__var_tracker.var_id, scope=scope
        )


class RouterMethod(BaseModel):
    fn: str
    args: Optional[list]



from __future__ import annotations
from typing import Any, Protocol, Union, cast
from typing_extensions import Self
from instaui.vars.types import TMaybeRef


class PropsProtocol(Protocol):
    def props(self, add: Union[str, dict[str, Any], TMaybeRef]) -> Self: ...


class CanDisabledMixin:
    def disabled(self, disabled: TMaybeRef[bool] = True) -> Self:
        return cast(PropsProtocol, self).props({"disabled": disabled})  # type: ignore



from fastapi import APIRouter
from instaui.handlers import watch_handler
from instaui.launch_collector import get_launch_collector

from . import _utils


def create_router(router: APIRouter):
    _async_handler(router)
    _sync_handler(router)


def _async_handler(router: APIRouter):
    @router.post(watch_handler.ASYNC_URL)
    async def _(data: dict):
        hkey = data.pop("key")
        handler_info = watch_handler.get_handler_info(hkey)
        if handler_info is None:
            return {"error": "watch handler not found"}

        _utils.update_app_page_info(data)

        result = await handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return _utils.response_web_data(handler_info.outputs_binding_count, result)


def _sync_handler(router: APIRouter):
    @router.post(watch_handler.SYNC_URL)
    def _(data: dict):
        hkey = data.pop("key")
        handler_info = watch_handler.get_handler_info(hkey)
        if handler_info is None:
            return {"error": "watch handler not found"}

        _utils.update_app_page_info(data)

        result = handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return _utils.response_web_data(handler_info.outputs_binding_count, result)

    if get_launch_collector().debug_mode:

        @router.get("/instaui/watch-infos", tags=["instaui-debug"])
        def watch_infos():
            return watch_handler.get_statistics_info()


def _get_binds_from_data(data: dict):
    return data.get("input", [])



from __future__ import annotations
from typing import Optional

from instaui.common.binding_track_mixin import is_binding_tracker
from instaui.runtime._index import get_app_slot

from .ref_base import RefBase


class UsePageTitleRef(RefBase):
    def __init__(
        self,
        deep_compare: bool = False,
        options: Optional[dict] = None,
    ) -> None:
        super().__init__(
            ref_type="usePageTitle", deep_compare=deep_compare, args=options
        )


def use_page_title(title: Optional[str] = None) -> str:
    """Set the title of the HTML document.

    Args:
        title (str): The title of the HTML document.
    """
    if not is_binding_tracker(title) and title is not None:
        get_app_slot()._html_resource.title = title

    return get_app_slot().base_scope.use_page_title_ref



from __future__ import annotations
from abc import abstractmethod
from typing import Any, Optional, Union, cast
from typing_extensions import Self
from dataclasses import dataclass, field
from instaui.common.binding_track_mixin import (
    is_binding_tracker,
    BindingTrackerMixin,
    mark_binding,
    try_mark_binding,
)
from instaui.common.var_track_mixin import VarTrackerMixin, mark_as_used
from instaui.runtime.scope import Scope
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.pathable import PathableMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin


@dataclass
class PathInfo:
    name: str
    args: Optional[list[Any]] = field(default=None)
    _is_binds: Optional[list[bool]] = field(default_factory=list, init=False)

    def __post_init__(self):
        self._is_binds = (
            None
            if self.args is None
            else [is_binding_tracker(arg) for arg in self.args]
        )

    def is_all_bind(self):
        return self._is_binds is not None and all(self._is_binds)

    def is_all_const(self):
        return self._is_binds is None or all(not bind for bind in self._is_binds)

    def mark_as_used(self):
        mark_as_used(self.args)

    def try_mark_binding(self, scope: Scope):
        if self.args is None:
            return [self.name]

        return [
            self.name,
            [try_mark_binding(arg, scope=scope) for arg in self.args],
            [int(bind) for bind in cast(list[bool], self._is_binds)],
        ]


class PathVar(PathableMixin):
    def __getitem__(self, item: Union[str, int]):
        return PathTrackerBindable(self)[item]

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def not_(self):
        return PathTrackerBindable(self).not_()

    def __add__(self, other: str):
        return PathTrackerBindable(self) + other

    def __radd__(self, other: str):
        return other + PathTrackerBindable(self)

    def __sub__(self, other: Union[int, float]):
        return PathTrackerBindable(self) - other

    def __rsub__(self, other: Union[int, float]):
        return other - PathTrackerBindable(self)

    def __mul__(self, other: Union[int, float]):
        return PathTrackerBindable(self) * other

    def __rmul__(self, other: Union[int, float]):
        return other * PathTrackerBindable(self)

    def __truediv__(self, other: Union[int, float]):
        return PathTrackerBindable(self) / other

    def __rtruediv__(self, other: Union[int, float]):
        return other / PathTrackerBindable(self)

    def __and__(self, other: Any):
        return other & PathTrackerBindable(self)

    def __or__(self, other: Any):
        return other | PathTrackerBindable(self)

    def __lt__(self, other):
        return PathTrackerBindable(self) < other

    def __le__(self, other):
        return PathTrackerBindable(self) <= other

    def __gt__(self, other):
        return PathTrackerBindable(self) > other

    def __ge__(self, other):
        return PathTrackerBindable(self) >= other

    def __ne__(self, other):
        return PathTrackerBindable(self) != other

    def len_(self):
        return PathTrackerBindable(self).len_()


class PathTracker(PathableMixin):
    def __init__(self, paths: Optional[list[PathInfo]] = None):
        self.paths = paths or []

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def __getitem__(self, key) -> Self:
        return self.__new_self__([*self.paths, PathInfo("bind", [key])])

    def __getattr__(self, key) -> Self:
        return self.__new_self__([*self.paths, PathInfo("bind", [key])])

    def not_(self) -> Self:
        return self.__new_self__([*self.paths, PathInfo("!")])

    def __add__(self, other: str) -> Self:
        return self.__new_self__([*self.paths, PathInfo("+", [other])])

    def __radd__(self, other: str) -> Self:
        return self.__new_self__([*self.paths, PathInfo("~+", [other])])

    def __sub__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("-", [other])])

    def __rsub__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("~-", [other])])

    def __mul__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("*", [other])])

    def __rmul__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("~*", [other])])

    def __truediv__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("/", [other])])

    def __rtruediv__(self, other: Union[int, float]):
        return self.__new_self__([*self.paths, PathInfo("~/", [other])])

    def __and__(self, other: Any):
        return self.__new_self__([*self.paths, PathInfo("&&", [other])])

    def __or__(self, other: Any):
        return self.__new_self__([*self.paths, PathInfo("||", [other])])

    def __lt__(self, other):
        return self.__new_self__([*self.paths, PathInfo("<", [other])])

    def __le__(self, other):
        return self.__new_self__([*self.paths, PathInfo("<=", [other])])

    def __gt__(self, other):
        return self.__new_self__([*self.paths, PathInfo(">", [other])])

    def __ge__(self, other):
        return self.__new_self__([*self.paths, PathInfo(">=", [other])])

    def __ne__(self, other):
        return self.__new_self__([*self.paths, PathInfo("!=", [other])])

    def len_(self):
        return self.__new_self__([*self.paths, PathInfo("len")])

    @abstractmethod
    def __new_self__(self, paths: list[PathInfo]) -> Self:
        pass


class PathTrackerBindable(
    PathTracker,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    ElementBindingMixin,
    StrFormatBindingMixin,
    BindingTrackerMixin,
    VarTrackerMixin,
):
    def __init__(self, source: PathableMixin):
        super().__init__()
        self.__source = source

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def __new_self__(self, paths: list[Union[str, list[str]]]) -> PathTrackerBindable:
        obj = PathTrackerBindable(self.__source)
        obj.paths = paths
        return obj

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    def _mark_as_used(self):
        cast(VarTrackerMixin, self.__source)._mark_as_used()

        for path in self.paths:
            path.mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        assert is_binding_tracker(self.__source)
        data = mark_binding(self.__source, scope=scope)

        if self.paths:
            data["path"] = _try_mark_binding_path_infos(self.paths, scope=scope)

        return data


def _try_mark_binding_path_infos(path_infos: list[PathInfo], scope: Scope):
    all_paths = all(path.name == "bind" for path in path_infos)
    all_const_paths = all_paths and all(path.is_all_const() for path in path_infos)

    if all_const_paths:
        return [cast(list, path.args)[0] for path in path_infos]

    return [path.try_mark_binding(scope=scope) for path in path_infos]



import typing

from instaui.common.binding_track_mixin import try_mark_binding
from instaui.common.var_track_mixin import mark_as_used
from . import _types
from . import _utils

from instaui.common.jsonable import Jsonable
from instaui.runtime import get_current_scope

from instaui.vars.mixin_types.py_binding import CanOutputMixin, outputs_to_config
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper


class JsWatch(Jsonable):
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
        immediate: bool = True,
        deep: typing.Union[bool, int] = False,
        once: bool = False,
        flush: typing.Optional[_types.TFlush] = None,
    ) -> None:
        inputs = observable_helper.auto_made_inputs_to_slient(inputs, outputs)

        self.code = code

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )
        self._outputs = [output for output in outputs or []]

        mark_as_used(self._inputs)
        mark_as_used(self._outputs)

        self._inputs = [try_mark_binding(input) for input in self._inputs]
        self._outputs = outputs_to_config(self._outputs)

        if immediate is not True:
            self.immediate = immediate

        if deep is not False:
            _utils.assert_deep(deep)
            self.deep = deep

        if once is not False:
            self.once = once

        if flush is not None:
            self.flush = flush

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._inputs:
            data["inputs"] = self._inputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        if self._outputs:
            data["outputs"] = self._outputs

        return data


def js_watch(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str,
    immediate: bool = True,
    deep: typing.Union[bool, int] = False,
    once: bool = False,
    flush: typing.Optional[_types.TFlush] = None,
):
    """
    Creates a client-side observer that executes JavaScript code in response to reactive source changes.

    Args:
        inputs (typing.Optional[typing.Sequence], optional): Reactive sources to observe. Changes to these sources
                                   trigger the watcher's JavaScript execution.
        outputs (typing.Optional[typing.Sequence], optional): Output targets associated with this watcher. Used for
                                    coordination with other observers.
        code (str, optional): JavaScript code to execute when changes are detected. The code has access
                  to the current values of observed inputs through the `args` parameter.
        immediate (bool, optional):If True, executes the watcher immediately after creation with current values. Defaults to True.
        deep (typing.Union[bool, int], optional): Controls depth of change detection:
                               - True: Recursively tracks nested properties
                               - False: Shallow comparison only
                               - int: Maximum depth level to track (for complex objects).
                               Defaults to False.
        once (bool, optional): If True, automatically stops observation after first trigger. Defaults to False.
        flush (typing.Optional[_types.TFlush], optional): Controls when to flush updates:
                                      - 'sync': Execute immediately on change
                                      - 'post': Batch updates and execute after current tick
                                      - 'pre': Execute before render phase (if applicable)

    # Example:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state('')
        ui.js_watch(inputs=[num], outputs=[msg], code="num => `The number is ${num}`")

        html.number(num)
        ui.text(msg)
    """

    watch = JsWatch(code, inputs, outputs, immediate, deep, once, flush)
    get_current_scope().register_js_watch(watch)
    return watch



from __future__ import annotations
from typing import Optional

from instaui.runtime._index import get_app_slot

from .ref_base import RefBase


class UseLanguageRef(RefBase):
    def __init__(
        self,
        deep_compare: bool = False,
        options: Optional[dict] = None,
    ) -> None:
        super().__init__(
            ref_type="useLanguage", deep_compare=deep_compare, args=options
        )


def use_language() -> str:
    """
    This function returns the current application's language setting, making it convenient to provide a unified language configuration for surrounding frameworks.
    """

    return get_app_slot().base_scope.use_language_ref



from __future__ import annotations
from typing import Optional
from fastapi import Request
from contextvars import ContextVar, Token

current_request_ctx: ContextVar[Optional[Request]] = ContextVar(
    "current_request", default=None
)


def set_current_request(request: Request):
    return current_request_ctx.set(request)


def reset_current_request(token: Token) -> None:
    current_request_ctx.reset(token)


def get_current_request() -> Request:
    return current_request_ctx.get()  # type: ignore



from __future__ import annotations
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Union
from typing_extensions import TypedDict, Unpack
import webview
import webview.http as http
from webview.guilib import GUIType

from instaui.runtime import get_app_slot, new_app_slot, reset_app_slot
from instaui.launch_collector import get_launch_collector
from instaui.runtime.utils import init_base_scope

from . import resource
from . import api
from . import _utils


class WebviewWrapper:
    """Example usage:
    .. code-block:: python
        from instaui import ui

        @ui.page("/")
        def index_page():
            ui.content("Hello, world!")

        ui.webview().run()
    """

    def __init__(
        self,
        *,
        assets_path: Union[str, Path] = "./webview_assets",
        debug: bool = False,
        auto_create_window: Union[bool, str] = "/",
        on_app_mounted: Optional[Callable] = None,
    ) -> None:
        """Create a new webview wrapper.

        Args:
            assets_path (Union[str, Path], optional): Path to store assets. Defaults to "./webview_assets".
            debug (bool, optional): Whether to run in debug mode. Defaults to False.
            auto_create_window (Union[bool, str], optional): Whether to create a window automatically. If a string is provided, it will be used as the initial page URL. Defaults to "/".
            testing (bool, optional): Whether to run in testing mode. Defaults to False.
        """

        get_launch_collector().setup(debug_mode=debug)

        self.assets_path = (
            Path(self._get_assets_path(assets_path))
            if isinstance(assets_path, str)
            else assets_path
        )
        _utils.reset_dir(self.assets_path)
        self.debug = debug
        self.on_app_mounted = on_app_mounted

        self._auto_create_window = auto_create_window

    def create_window(
        self,
        page_url: str = "/",
    ):
        """Create a new window. Returns the window object of pywebview.

        Args:
            page_url (str, optional): Page URL to load. Defaults to "/".

        """
        launch_collector = get_launch_collector()
        with _scope():
            app = get_app_slot()
            app._page_path = page_url
            page_info = launch_collector._page_router[page_url]

            init_lifespans = [
                iter(event()) for event in launch_collector.page_init_lifespans
            ]
            for event in init_lifespans:
                next(event)

            page_info.func()

            for event in init_lifespans:
                try:
                    next(event)
                except StopIteration:
                    pass

            resource_info = resource.resource_to_assets(
                page_url=page_url,
                assets_path=self.assets_path,
                on_app_mounted=self.on_app_mounted,
            )

        window = webview.create_window(
            resource_info.title, resource_info.index_html_url, js_api=api.Api()
        )

        if self.on_app_mounted:

            def on_app_mounted():
                self.on_app_mounted(window)  # type: ignore

            window.expose(on_app_mounted)

        return window

    def run(self, **webview_start_args: Unpack[WebviewStartArgs]):
        """Run the webview.

        Args:
            :param func: Function to invoke upon starting the GUI loop.
            :param args: Function arguments. Can be either a single value or a tuple of
                values.
            :param localization: A dictionary with localized strings. Default strings
                and their keys are defined in localization.py.
            :param gui: Force a specific GUI. Allowed values are ``cef``, ``qt``,
                ``gtk``, ``mshtml`` or ``edgechromium`` depending on a platform.
            :param http_server: Enable built-in HTTP server. If enabled, local files
                will be served using a local HTTP server on a random port. For each
                window, a separate HTTP server is spawned. This option is ignored for
                non-local URLs.
            :param user_agent: Change user agent string.
            :param private_mode: Enable private mode. In private mode, cookies and local storage are not preserved.
                Default is True.
            :param storage_path: Custom location for cookies and other website data
            :param menu: List of menus to be included in the app menu
            :param server: Server class. Defaults to BottleServer
            :param server_args: dictionary of arguments to pass through to the server instantiation
            :param ssl: Enable SSL for local HTTP server. Default is False.
            :param icon: Path to the icon file. Supported only on GTK/QT.
        """

        if self._auto_create_window:
            self.create_window(
                "/" if self._auto_create_window is True else self._auto_create_window
            )

        webview.start(**webview_start_args, debug=self.debug)

    @staticmethod
    def _get_assets_path(file: Union[str, Path]) -> Path:
        if isinstance(file, str):
            import inspect

            frame = inspect.currentframe().f_back.f_back.f_back.f_back.f_back  # type: ignore
            assert frame is not None
            script_file = inspect.getfile(frame)
            file = Path(script_file).parent.joinpath(file)

        return file


@contextmanager
def _scope():
    app, token = new_app_slot("webview")
    init_base_scope(app)
    yield
    reset_app_slot(token)


class WebviewStartArgs(TypedDict, total=False):
    func: Union[Callable[..., None], None]
    args: Union[Iterable[Any], None]
    localization: dict[str, str]
    gui: Union[GUIType, None]
    http_server: bool
    http_port: Union[int, None]
    user_agent: Union[str, None]
    private_mode: bool
    storage_path: Union[str, None]
    menu: list[Any]
    server: type[http.ServerType]  # type: ignore
    server_args: dict[Any, Any]
    ssl: bool
    icon: Union[str, None]



from typing import cast
from instaui.common.binding_track_mixin import is_binding_tracker
from instaui.common.var_track_mixin import mark_as_used
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.vue_computed import VueComputed


def str_format(template: str, *args, **kwargs):
    """
    Formats a string template with positional and keyword arguments.

    Args:
        template (str): The string template containing {} placeholders for formatting.
        *args: Variable length positional arguments for positional formatting.
        **kwargs: Arbitrary keyword arguments for named formatting.

    Example:
    .. code-block:: python
        # Positional formatting
        ui.str_format("pos:a={},b={}", a, b)

        # Index-based positional formatting
        ui.str_format("num pos:a={0},b={1}", a, b)

        # Named formatting
        ui.str_format("name pos:b={b},a={a}", a=a, b=b)
    """
    bindings = {}
    tran_args = []

    mark_as_used(args)
    mark_as_used(kwargs)

    for idx, arg in enumerate(args):
        is_mixin = is_binding_tracker(arg)
        value = (
            cast(StrFormatBindingMixin, arg)._to_str_format_binding(idx)
            if is_mixin
            else arg
        )
        tran_args.append(value[-1] if is_mixin else value)
        if is_mixin:
            bindings[value[0]] = arg

    tran_kwargs = {}

    for idx, (k, v) in enumerate(kwargs.items()):
        is_mixin = is_binding_tracker(v)
        value = (
            cast(StrFormatBindingMixin, v)._to_str_format_binding(idx)
            if is_mixin
            else v
        )
        tran_kwargs[k] = value[-1] if is_mixin else value
        if is_mixin:
            bindings[value[0]] = v

    code = "()=>`" + template.format(*tran_args, **tran_kwargs) + "`"
    return VueComputed(code, bindings=bindings)



from __future__ import annotations
from typing import TypeVar, cast

from .ref_base import RefBase
from .state import RefProxy

_T = TypeVar("_T")


class SessionStorageRef(RefBase):
    def __init__(
        self,
        key: str,
        value,
        deep_compare: bool = False,
    ) -> None:
        super().__init__(
            ref_type="storage",
            deep_compare=deep_compare,
            args={"type": "session", "key": key, "value": value},
        )


def session_storage(key: str, value: _T, deep_compare: bool = False) -> _T:
    """
    Creates a reactive state object synchronized with the browser's session storage.

    This function initializes a reactive value tied to a given key in session storage.
    The value is preserved across page reloads during the same browser session, but
    will be cleared once the tab or window is closed.

    Args:
        key (str): The session storage key to associate with the value.
        value (_T): The default value to use if no value exists in session storage.

    Returns:
        _T: A reactive value linked to the specified session storage key.

    Example:
    .. code-block:: python

        from instaui import ui, html

        @ui.page('/')
        def index():
            name = ui.session_storage("username", "")
            html.input(name)
    """

    if isinstance(value, RefProxy):
        return value

    return cast(_T, SessionStorageRef(key, value=value, deep_compare=deep_compare))



from __future__ import annotations
import threading
from typing import Callable, Optional, Sequence, TYPE_CHECKING
import ast
from . import _utils
from instaui.vars.event_context import DatasetEventContext
from instaui.runtime import update_web_server_info

if TYPE_CHECKING:
    from instaui.vars.mixin_types.py_binding import CanOutputMixin, CanInputMixin

ASYNC_URL = "/instaui/event/async"
SYNC_URL = "/instaui/event/sync"
update_web_server_info(event_url=SYNC_URL, event_async_url=ASYNC_URL)
_event_handlers: dict[str, _utils.HandlerInfo] = {}
dict_lock = threading.Lock()


class EventDataSetTypeAdapter:
    def to_python_value(self, value, *args, **kwargs):
        if value is None:
            return None
        return ast.literal_eval(value)


def register_event_handler(
    key: str,
    handler: Callable,
    outputs_binding: Optional[Sequence[CanOutputMixin]],
    inputs_binding: Optional[Sequence[CanInputMixin]],
):
    if key in _event_handlers:
        return

    custom_type_adapter_map = {
        i: EventDataSetTypeAdapter()
        for i, binding in enumerate(inputs_binding or [])
        if isinstance(binding, DatasetEventContext)
    }

    handler_info = _utils.HandlerInfo.from_handler(
        handler,
        len(list(outputs_binding)) if outputs_binding else 0,
        custom_type_adapter_map=custom_type_adapter_map,
    )

    with dict_lock:
        _event_handlers[key] = handler_info


def get_handler(key: str) -> Optional[_utils.HandlerInfo]:
    return _event_handlers.get(key)


def get_statistics_info():
    return {
        "_event_handlers count": len(_event_handlers),
        "_event_handlers keys": list(_event_handlers.keys()),
    }


create_handler_key = _utils.create_handler_key



import typing
from instaui.event.event_modifier import TEventModifier
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars._types import InputBindingType
from instaui.common.var_track_mixin import mark_as_used
from instaui.common.binding_track_mixin import try_mark_binding
from .event_mixin import EventMixin


class VueEvent(EventMixin):
    """
    Create an event object that can be bound to a UI component's event listener.

    This function generates a callable event handler with optional contextual bindings.
    The event logic is defined via a code string, which can reference bound variables.

    Args:
        code (str): A string containing the executable logic for the event handler.
                    Typically contains a function body or expression that utilizes bound variables.
        bindings (typing.Optional[dict[str, typing.Any]], optional): A dictionary mapping variable names to values that should be available in the
            event handler's context. If None, no additional bindings are created.. Defaults to None.

    Example:
    .. code-block:: python
        a = ui.state(1)

        event = ui.vue_event(bindings={"a": a}, code=r'''()=> { a.value +=1}''')

        html.span(a)
        html.button("plus").on("click", event)
    """

    def __init__(
        self,
        *,
        code: str,
        bindings: typing.Optional[dict[str, typing.Any]] = None,
    ):
        self._code = code
        self._bindings = bindings

    def event_type(self):
        return "vue"

    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]],
    ) -> dict:
        data = {}
        data["code"] = self._code
        data["type"] = self.event_type()

        mark_as_used(self._bindings)

        if self._bindings:
            data["inputs"] = {
                name: {
                    "value": try_mark_binding(v),
                    "type": v._to_event_input_type().value
                    if isinstance(v, CanInputMixin)
                    else InputBindingType.Data,
                }
                for name, v in self._bindings.items()
            }
        return data


vue_event = VueEvent



from ._index import get_current_scope, get_app_slot
from .scope import Scope


def try_new_scope_on_ui_state() -> Scope:
    scope = get_current_scope()
    vfor = scope.get_running_vfor()

    if vfor is None:
        return scope

    app = get_app_slot()
    _new_scope = Scope(app.gen_scope_id())

    _new_scope.__enter__()

    @vfor._on_exit
    def _():
        _new_scope.__exit__()

    return _new_scope



from __future__ import annotations
from typing import (
    Any,
    Optional,
    Union,
)

from instaui.components.element import Element
from instaui.components.value_element import ValueElement


from instaui.vars.types import TMaybeRef
from ._mixins import InputEventMixin


class Checkbox(InputEventMixin, ValueElement[Union[bool, str]]):
    def __init__(
        self,
        value: Union[Union[bool, str], TMaybeRef[Union[bool, str]], None] = None,
        *,
        model_value: Optional[TMaybeRef[Union[bool, str]]] = None,
        checked: Optional[TMaybeRef[bool]] = None,
        id: Optional[Any] = None,
    ):
        super().__init__("input", value, is_html_component=True)
        self.props({"type": "checkbox"})
        if id is not None:
            self.props({"id": id})
        if checked is not None:
            self.props({"checked": checked})
        if model_value is not None:
            self.props({"value": model_value})

    def _input_event_mixin_element(self) -> Element:
        return self



from __future__ import annotations
from typing import (
    Generic,
    Optional,
    Union,
    TypeVar,
)
from instaui.common.binding_track_mixin import try_mark_binding
from instaui.common.var_track_mixin import mark_as_used
from instaui.components.element import Element
from instaui import consts

from instaui.vars.types import TMaybeRef
from instaui.vars.web_computed import WebComputed
from instaui.ui_functions.ui_types import is_bindable


_T = TypeVar("_T")


class ValueElement(Element, Generic[_T]):
    def __init__(
        self,
        tag: Optional[str] = None,
        value: Union[_T, TMaybeRef[_T], None] = None,
        is_html_component: bool = False,
        value_name: str = "value",
    ):
        super().__init__(tag)
        self.__is_html_component = is_html_component

        if value is not None:
            if is_bindable(value):
                if isinstance(value, WebComputed):
                    self.props({value_name: value})
                else:
                    self.vmodel(value, prop_name=value_name)
            else:
                self.props({value_name: value})

    def vmodel(
        self,
        value,
        modifiers: Union[consts.TModifier, list[consts.TModifier], None] = None,
        *,
        prop_name: str = "value",
    ):
        mark_as_used(value)
        return super().vmodel(
            try_mark_binding(value),
            modifiers,
            prop_name=prop_name,
            is_html_component=self.__is_html_component,
        )



from typing import Optional
from typing_extensions import Unpack, TypedDict
from instaui.consts import _T_App_Mode
from contextvars import ContextVar, Token

from instaui.launch_collector import get_launch_collector
from ._app import App, DefaultApp


def _create_default():
    if App._default_app_slot is None:
        App._default_app_slot = DefaultApp(mode="web")
    return App._default_app_slot


_app_var: ContextVar[App] = ContextVar("_app_var", default=_create_default())


def use_default_app_slot():
    assert App._default_app_slot is not None, "Default app slot is not set"
    _app_var.set(App._default_app_slot)


def get_default_app_slot():
    return _create_default()


def get_app_slot() -> App:
    return _app_var.get()


def get_current_scope():
    current_scope = get_app_slot()._scope_stack[-1]
    if current_scope is None:
        raise ValueError("No current scope")
    return current_scope


def get_slot_stacks():
    return get_app_slot()._slots_stacks


def pop_slot():
    get_slot_stacks().pop()


def new_app_slot(mode: _T_App_Mode, *, app_meta: Optional[dict] = None):
    app = App(mode=mode, meta=app_meta, debug_mode=get_launch_collector().debug_mode)
    token = _app_var.set(app)
    return app, token


def reset_app_slot(token: Token[App]):
    _app_var.reset(token)


def in_default_app_slot():
    return isinstance(get_app_slot(), DefaultApp)


def check_default_app_slot_or_error(
    error_message="Operations are not allowed outside of ui.page",
):
    if isinstance(get_app_slot(), DefaultApp):
        raise ValueError(error_message)


class TWebServerInfo(TypedDict, total=False):
    watch_url: Optional[str]
    watch_async_url: Optional[str]
    event_url: Optional[str]
    event_async_url: Optional[str]


def update_web_server_info(**kwargs: Unpack[TWebServerInfo]):
    if App._web_server_info is None:
        App._web_server_info = {}

    data = {k: v for k, v in kwargs.items() if v is not None}

    App._web_server_info.update(data)



from __future__ import annotations
import typing
from instaui import html
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef


class RouterLink(Element):
    def __init__(self, text: TMaybeRef[str], *, to: str):
        """
        Represents a navigational link element in a SPA (Single Page Application) that allows users
        to navigate between different routes without a full page reload.

        Args:
            text (TMaybeRef[str]): The display text of the link. Can be a static string or a reactive reference.
            to (str): The target route path that the link navigates to when clicked.


        Example:
        .. code-block:: python
            from instaui import html, spa_router, ui

            @ui.page()
            def index():
                # Basic route registration
                def home():
                    html.paragraph("hello home")

                def user_page():
                    name = spa_router.get_params("name")
                    html.paragraph(ui.str_format("welcome: {name}", name=name))

                # Add routes with different configurations
                spa_router.add_route(home)
                spa_router.add_route(user_page, path="/user/:name")

                # ui
                with ui.row():
                    spa_router.link("home page", to="/")
                    spa_router.link("user page", to="/user/foo")
                    spa_router.link("user page", to="/user/bar")
        """

        super().__init__("router-link")

        self.props({"to": to})

        if text is not None:
            with self.add_slot("default"):
                html.span(text)

    @classmethod
    def by_name(
        cls,
        text: TMaybeRef[str],
        *,
        name: str,
        params: typing.Optional[dict[str, typing.Any]] = None,
    ) -> RouterLink:
        to: dict = {"name": name}
        if params:
            to["params"] = params

        return cls(text, to=to)  # type: ignore


class RouterView(Element):
    def __init__(self):
        """
        Container element that displays the current route component in SPA routing.

        The router view acts as a dynamic container that automatically renders the component
        associated with the current route. It updates when navigation occurs via links or
        programmatic routing, providing the main content area for single-page applications.

        Example:
        .. code-block:: python
            from instaui import html, spa_router, ui

            @ui.page()
            def index():
                # Basic route registration
                def home():
                    html.paragraph("hello home")

                def user_page():
                    name = spa_router.get_params("name")
                    html.paragraph(ui.str_format("welcome: {name}", name=name))

                # Add routes with different configurations
                spa_router.add_route(home)
                spa_router.add_route(user_page, path="/user/:name")

                # ui
                with ui.row():
                    spa_router.link("home page", to="/")
                    spa_router.link("user page", to="/user/foo")
                    spa_router.link("user page", to="/user/bar")
        """
        super().__init__("router-view")



from __future__ import annotations
from typing import Any
from instaui.common.binding_track_mixin import is_binding_tracker, try_mark_binding
from instaui.common.var_track_mixin import mark_as_used
from instaui.vars.types import TMaybeRef
from instaui.components.logic_component import LogicComponent


class Content(LogicComponent):
    def __init__(self, content: TMaybeRef[Any]):
        """Content to be displayed on the page, typically used for pure text content within slots.

        Args:
            content (TMaybeRef[Any]): The textual content to display.

        Examples:
        .. code-block:: python
            with html.div():
                ui.content("Hello, world!")
        """
        super().__init__("content")
        mark_as_used(content)
        self._content = try_mark_binding(content)
        self._is_binding = is_binding_tracker(content)

    def _to_json_dict(self) -> dict:
        data = super()._to_json_dict()

        if self._is_binding:
            data["r"] = 1

        if self._content:
            data["value"] = self._content

        return data



from typing import Any
from instaui.common.binding_track_mixin import mark_binding, is_binding_tracker
from instaui.common.var_track_mixin import mark_as_used, is_var_tracker


def convert_reference(binding: Any):
    """
    Allows reactive variable configurations to be passed into custom components, enabling them to obtain their Ref references.

    Args:
        binding (Any): The reactive binding or variable to be converted

    Examples:
    .. code-block:: python

        class CustomElement(custom.element, esm="./custom_element.js"):
            def __init__(self, ref):
                super().__init__()
                self.props({"ref_binding": custom.convert_reference(ref)})
    """

    assert is_var_tracker(binding), "binding should be a VarTrackerMixin"
    mark_as_used(binding)

    assert is_binding_tracker(binding), "binding should be a BindingTrackerMixin"
    config = mark_binding(binding)
    return config



__all__ = [
    "url_location",
    "set_cookie",
    "cookie_output",
    "cookie_input",
]


from .url_location import UrlLocation as url_location
from .cookie import cookie_input, cookie_output, set_cookie



from __future__ import annotations
from typing import TYPE_CHECKING
from instaui.common.binding_track_mixin import BindingTrackerHelper, BindingTrackerMixin
from instaui.common.jsonable import Jsonable
from instaui.common.var_track_mixin import VarTrackerHelper, VarTrackerMixin
from instaui.runtime import try_new_scope_on_ui_state
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.py_binding import CanOutputMixin, CanInputMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin

if TYPE_CHECKING:
    from instaui.runtime.scope import Scope


class ElementRef(
    Jsonable,
    CanOutputMixin,
    CanInputMixin,
    ElementBindingMixin,
    BindingTrackerMixin,
    ObservableMixin,
    VarTrackerMixin,
):
    def __init__(self) -> None:
        """
        A reference holder for a frontend DOM element, enabling two-way interaction between
        Python and JavaScript. The referenced element can be used as input/output in JS
        events, and methods can be executed dynamically on the element at runtime.

        Args:
            None: The reference is created without initialization parameters. It will be
                bound to an element using ``element_ref(...)`` during UI construction.

        Example:
        .. code-block:: python
            # 1️ Use ElementRef inside ui.event as an output
            cp = ui.element_ref()

            @ui.event(outputs=[cp])
            def on_click():
                return ui.run_element_method("reset")

            Counter("my counter").element_ref(cp)
            html.button("reset").on_click(on_click)


            # 2️⃣ Use ElementRef as input to ui.js_event
            ele = ui.element_ref()
            click_event = ui.js_event(inputs=[ele], code="(ele)=> ele.increment()")

            CounterWithRefInput().element_ref(ele)
            html.button("click").on_click(click_event)


            # 3️⃣ Access DOM element inside JavaScript event using `ele.$el ?? ele`
            ele_dom = ui.element_ref()
            output = ui.state("")

            ui.text("custom").element_ref(ele_dom).on(
                "click",
                ui.js_event(
                    inputs=[ele_dom],
                    outputs=[output],
                    code="ele => (ele.$el ?? ele).tagName"
                ),
            )
            ui.text(output)
        """

        self._define_scope = try_new_scope_on_ui_state()

        self.__var_tracker = VarTrackerHelper(
            var_id_gen_fn=lambda: self._define_scope.register_element_ref(self)
        )
        self.__binding_tracker = BindingTrackerHelper(define_scope=self._define_scope)

    def __to_binding_config(
        self,
    ):
        raise NotImplementedError()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__var_tracker.var_id

        return data

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.ElementRefAction

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.ElementRef

    def _mark_as_used(self):
        self.__var_tracker.mark_as_used()

    def _mark_binding(self, scope: Scope) -> dict:
        return self.__binding_tracker.mark_binding(
            var_id=self.__var_tracker.var_id, scope=scope
        )


def run_element_method(method_name: str, *args, **kwargs):
    return {
        "method": method_name,
        "args": args,
    }



from pathlib import Path
from typing import Literal

_THIS_DIR = Path(__file__).parent
_TEMPLATES_DIR = _THIS_DIR.joinpath("templates")
_STATIC_DIR = _THIS_DIR.joinpath("static")
_COMPILED_DIR = _STATIC_DIR.joinpath("compiled")

INDEX_TEMPLATE_PATH = _TEMPLATES_DIR.joinpath("index.html")
FAVICON_PATH = _STATIC_DIR.joinpath("insta-ui.ico")

# compiled files
APP_ES_JS_PATH = _COMPILED_DIR.joinpath("insta-ui.esm-browser.prod.js")
APP_ES_JS_MAP_PATH = _COMPILED_DIR.joinpath("insta-ui.js.map")
APP_CSS_PATH = _COMPILED_DIR.joinpath("insta-ui.css")
VUE_ES_JS_PATH = _COMPILED_DIR.joinpath("vue.esm-browser.prod.js")
# tools
TOOLS_BROWSER_JS_PATH = _COMPILED_DIR.joinpath("instaui-tools-browser.js")

PAGE_TITLE = "insta-ui"
SCOPED_STYLE_GROUP_ID = "insta-scoped-style"

_T_App_Mode = Literal["zero", "web", "webview"]
TModifier = Literal["trim", "number", "lazy"]



from typing import Any, Optional, Sequence, Union
from instaui.common.binding_track_mixin import mark_binding, try_mark_binding
from instaui.common.jsonable import Jsonable
from instaui.missing import TMissing
from instaui.vars.mixin_types.py_binding import CanOutputMixin


def _check_args(config: dict):
    for key in config.keys():
        if not isinstance(key, CanOutputMixin):
            raise TypeError(f"key {key} is not a CanOutputMixin")


def extract_bindings(pre_setup: list[list]):
    result = []
    for target, running_value, reset_value in pre_setup:
        result.append(target)
        if isinstance(running_value, PreSetupAction):
            result.extend(running_value._inputs or [])
        if isinstance(reset_value, PreSetupAction):
            result.extend(reset_value._inputs or [])

    return result


def convert_list2list(pre_setup: Optional[list]) -> list:
    if not pre_setup:
        return []

    first = pre_setup[0]
    # e.g [ref, True, False]
    if not isinstance(first, list):
        return [pre_setup]

    # e.g [[ref, True, False], [ref2, False, True]]
    return pre_setup


def convert_config(configs: list[list]):
    return [
        [
            mark_binding(target),
            _to_config_value(running_value),
            _to_config_value(reset_value),
        ]
        for target, running_value, reset_value in configs
    ]


def _to_config_value(maybe_action: Union[Any, TMissing]) -> Any:
    if isinstance(maybe_action, TMissing):
        return {
            "type": "missing",
        }

    if isinstance(maybe_action, PreSetupAction):
        return {
            "type": "action",
            "value": maybe_action._to_json_dict(),
        }

    return {"type": "const", "value": maybe_action}


class PreSetupAction(Jsonable):
    def __init__(
        self,
        *,
        code: str,
        inputs: Optional[Sequence] = None,
    ):
        self.type = "action"
        self._inputs = inputs
        self.code = code

    def _to_json_dict(self):
        data = super()._to_json_dict()
        if self._inputs:
            data["inputs"] = [try_mark_binding(binding) for binding in self._inputs]

        return data



from __future__ import annotations
from typing import Literal, Optional
from instaui.components.element import Element


class Link(Element):
    def __init__(
        self,
        href: Optional[str] = None,
        *,
        text: Optional[str] = None,
        target: Optional[
            Literal["_blank", "_self", "_parent", "_top", "_unfencedTop"]
        ] = None,
        type: Optional[str] = None,
    ):
        """
        A wrapper component for HTML anchor links with predefined styles and common attributes.

        This component simplifies the creation of styled links and supports both static and
        reactive properties for dynamic behavior.

        Args:
            href (Optional[str]): The URL the link points to. Corresponds to the 'href' attribute.
            text (Optional[str]): The display text of the link. If not set, child elements are used.
            target (Optional[Literal]): Specifies where to open the linked document
                                       (e.g., '_blank' for new tab).
            type (Optional[str]): Hints at the linked resource's MIME type (e.g., 'text/html').

        # Example:
        .. code-block:: python
            from instaui import ui

            # Basic link
            ui.link("https://example.com", text="Visit Example")

            # Link with custom target
            ui.link("https://example.com", text="Open in new tab", target="_blank")

            # Link with child content
            with ui.link("https://example.com"):
                ui.text("Styled link content")

        """
        super().__init__("ui-link")

        self.props(
            {
                "href": href,
                "text": text,
                "target": target,
                "type": type,
            }
        )



from __future__ import annotations
import threading
from collections.abc import Hashable
from typing import (
    Callable,
    Generic,
    Optional,
    Sequence,
    TypeVar,
)
from dataclasses import dataclass

from instaui.launch_collector import get_launch_collector
from instaui.runtime import update_web_server_info
from instaui.systems import func_system
from . import _utils


ASYNC_URL = "/instaui/watch/async"
SYNC_URL = "/instaui/watch/sync"

update_web_server_info(watch_url=SYNC_URL, watch_async_url=ASYNC_URL)

_watch_handlers: dict[Hashable, _utils.HandlerInfo] = {}
dict_lock = threading.Lock()


def register_handler(
    key: str,
    handler: Callable,
    outputs_binding_count: int,
):
    if key in _watch_handlers:
        return
    with dict_lock:
        _watch_handlers[key] = _utils.HandlerInfo.from_handler(
            handler, outputs_binding_count
        )


def get_handler_info(key: str) -> _utils.HandlerInfo:
    return _watch_handlers.get(key)  # type: ignore


def get_statistics_info():
    return {
        "_watch_handlers count": len(_watch_handlers),
        "_watch_handlers keys": list(_watch_handlers.keys()),
    }


def create_handler_key(
    page_path: str,
    handler: Callable,
    extra_key: Optional[Sequence[Hashable]] = None,
):
    _, lineno, _ = func_system.get_function_location_info(handler)
    key = (
        f"path:{page_path}|line:{lineno}"
        if get_launch_collector().debug_mode
        else f"{page_path}|{lineno}"
    )
    if extra_key:
        key = repr(tuple(extra_key) + (key,))

    return key


_TWatchStateValue = TypeVar("_TWatchStateValue")


@dataclass(frozen=True)
class WatchState(Generic[_TWatchStateValue]):
    new_value: _TWatchStateValue
    old_value: _TWatchStateValue
    modified: bool



from __future__ import annotations
from typing import Callable
from dataclasses import dataclass
from urllib.parse import quote


@dataclass
class PageInfo:
    path: str
    func: Callable

    def create_key(self) -> str:
        return quote(self.path)



from __future__ import annotations
from instaui.common.jsonable import Jsonable
from instaui.components.component import Component
from instaui.runtime.scope import Scope

from . import _types
from ._route_model import RouteCollector


class RouterBox_(Jsonable):
    def __init__(self):
        self._mode: _types.TRouterHistoryMode = "web"
        self._keep_alive: bool = False
        self._component_map: dict[str, dict] = {}
        self._route_collector: RouteCollector = RouteCollector()

    def add_component(
        self,
        path: str,
        components: list[Component],
        scope: Scope,
        lazy_loading: bool = False,
    ):
        config = {"items": components, "scope": scope}
        if lazy_loading:
            config["lazy"] = True

        self._component_map[path] = config

    def _to_json_dict(self):
        data = {}

        data["routes"] = self._route_collector._routes
        data["mode"] = self._mode

        if self._keep_alive is not False:
            data["kAlive"] = self._keep_alive

        return data



from __future__ import annotations

import typing
from instaui.common.binding_track_mixin import try_mark_binding, is_binding_tracker
from instaui.common.component_mixin import ComponentMixin
from instaui.common.var_track_mixin import mark_as_used
from instaui.vars.types import TMaybeRef
from instaui.components.logic_component import LogicContainer


class Match(LogicContainer):
    def __init__(self, cond: TMaybeRef[typing.Any]):
        """
        Creates a conditional logic container that renders different UI blocks
        based on the value of the given reactive reference. Each case is checked
        sequentially, and the first match will be displayed. A default case can be
        defined to handle unmatched values.

        Args:
            cond (TMaybeRef[typing.Any]): A reactive reference or plain value whose
                state determines which case block should be rendered.

        Example:
        .. code-block:: python
            a = ui.state("")

            html.input().vmodel(a)

            with ui.match(a) as mt:
                with mt.case("page1"):
                    html.paragraph("in page1 case")

                with mt.case("page2"):
                    html.paragraph("in page2 case")

                with mt.default():
                    html.paragraph("in default case")
        """

        super().__init__("match")
        mark_as_used(cond)

        self._cond = try_mark_binding(cond)
        self._const_cond = not is_binding_tracker(cond)
        self._cases: list[MatchCase] = []
        self._default_case: typing.Optional[DefaultCase] = None

        self._items: list[ComponentMixin] = []

    def __enter__(self):
        super().__enter__()
        return MatchWrapper(self)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["cond"] = self._cond
        if self._const_cond:
            data["const"] = 1

        if self._cases:
            data["cases"] = self._cases

        if self._default_case:
            data["default"] = self._default_case

        return data

    def add_item(self, item: ComponentMixin):
        pass


class MatchWrapper:
    def __init__(self, host: Match) -> None:
        self.__host = host

    def case(self, value: typing.Any):
        assert not is_binding_tracker(value), "Cannot use binding in case value"
        case = MatchCase(value)
        self.__host._cases.append(case)
        return case

    def default(self):
        default_case = DefaultCase()
        self.__host._default_case = default_case
        return default_case


class MatchCase(LogicContainer):
    def __init__(self, value: typing.Any):
        super().__init__("match-case")
        self._value = value
        self._items: list[ComponentMixin] = []

    def _to_json_dict(self) -> dict:
        data = {}
        data["value"] = self._value
        if self._items:
            data["items"] = self._items
        return data

    def add_item(self, item: ComponentMixin):
        self._items.append(item)


class DefaultCase(LogicContainer):
    def __init__(self):
        super().__init__("default-case")
        self._items: list[ComponentMixin] = []

    def _to_json_dict(self) -> dict:
        data = {}
        if self._items:
            data["items"] = self._items
        return data

    def add_item(self, item: ComponentMixin):
        self._items.append(item)
